from yostlabs.tss3.api import ThreespaceSensor, StreamableCommands, ThreespaceCmdResult, threespaceGetHeaderLabels

from enum import Enum
from typing import Any, Callable
from dataclasses import dataclass, field

class ThreespaceStreamingStatus(Enum):
    Data = 0        #Normal data update
    DataEnd = 1     #This is the last packet being sent for this data update. This allows the user to more efficently handle their callback.
                    #For example, if you have an expensive computation that needs done over all the data, but only once per frame, it would
                    #be preferable to buffer the data received via the callback and only do the computation when DataEnd is received
    Paused = 2      #Streaming has been paused
    Resumed = 3     #Streaming has been resumed

    #Streaming manager is resetting. It is required that the callback unregisters everything is has registered
    #This option is intended for shutdown purposes or in complex applications where the user needs to completely
    #disable the streaming manager for some reason
    Reset = 4

from typing import NamedTuple
ThreespaceStreamingOption = NamedTuple("ThreespaceStreamingOption", [("cmd", StreamableCommands), ("param", int|None)])
class ThreespaceStreamingManager:
    """
    A class that manages multiple clients wanting streamed data. Will update the streaming
    slots and speed dynamically based on given requirements and allow those clients to
    access that data without having to worry about which streaming slot is being used by what data.
    """ 

    @dataclass
    class Command:
        cmd: StreamableCommands = None
        param: int = None

        slot: int = None
        registrations: set = field(default_factory=set, init=False)

        active: bool = False #If not active then it must have been queued for addition, so it will be set active immediately

        labels: str = None

    @dataclass
    class Callback:
        func: Callable[[ThreespaceStreamingStatus],None] = None
        hz: int = None

        only_newest: bool = False

        @property
        def interval(self):
            if self.hz is None: return None
            return 1000000 // self.hz

    def __init__(self, sensor: ThreespaceSensor):
        self.sensor = sensor

        self.num_slots = len(self.get_slots_from_sensor()) #This is just so if number of available slots ever changes, this changes to match it
        self.registered_commands: dict[tuple, ThreespaceStreamingManager.Command] = {}
        self.slots: list[ThreespaceStreamingManager.Command] = [] #The same as self.registered_commands, but allow indexing based on slot instead

        self.last_response: ThreespaceCmdResult|None = None
        self.results: dict[tuple,Any] = {}

        self.callbacks: dict[Callable, ThreespaceStreamingManager.Callback] = {}

        #Objects currently pausing the streaming
        self.pausers: set[object] = set()
        self.lockers: set[object] = set()

        #Keeps track of how many packets have been read. Useful for consumers to know if the values have been updated since they last read
        self.sample_count = 0

        self.enabled = False
        self.is_streaming = False #Store this seperately to attempt to allow using both the regular streaming and streaming manager via pausing and such

        #Set the initial streaming speed
        self.interval = int(self.sensor.get_settings("stream_interval"))    

        self.dirty = False
        self.validated = False

        #Control variable to manually control when updating happens here
        self.block_updates = False

        #Using interval instead of HZ because more precise and the result of ?stream_hz may not be exactly equal to what is set
        #However the functions for interfacing with these are still done in Hz
        self.max_interval = 0xFFFFFFFF
        self.min_interval = 1000000 / 2000

    @property
    def paused(self):
        return len(self.pausers) > 0
    
    @property
    def locked(self):
        return len(self.lockers) > 0    

    def pause(self, locker: object):
        if locker in self.pausers: return True
        if self.locked: return False
        self.pausers.add(locker)
        if len(self.pausers) == 1 and self.is_streaming:
            self.__stop_streaming()
            for callback in self.callbacks:
                callback(ThreespaceStreamingStatus.Paused)
        return True

    def resume(self, locker: object):
        try:
            self.pausers.remove(locker)
        except KeyError:
            return

        #Attempt to start again
        if len(self.pausers) == 0:
            for callback in self.callbacks:
                callback(ThreespaceStreamingStatus.Resumed)
            self.__apply_streaming_settings_and_update_state()

    def lock_modifications(self, locker: object):
        """
        This still allows the streaming manager to operate and register new objects. However, registration
        is limited to commands and speeds that are already operatable. Essentially, after this is called,
        it is not possible to do actions that require updating the sensors onboard settings/state. This gurantees
        streaming will not be stopped/restarted for time sensitive applications. 
        Note: This INCLUDES pausing/resuming, enabling/disabling...
        If you need to lock modifications, then pause or resume. The locker should unlock modifications, call the necessary function, and then lock again
        """
        self.lockers.add(locker)
    
    def unlock_modifications(self, locker: object):
        if not locker in self.lockers: return
        self.lockers.remove(locker)
        if not self.locked and self.dirty:
            self.__apply_streaming_settings_and_update_state()

    def reset(self):
        #Prevent the callbacks unregistrations from instantly taking effect regardless of if they pass immediate_update or not
        self.block_updates = True
        values = list(self.callbacks.values()) #To prevent concurrent dict modification, cache this
        for cb in values:
            cb.func(ThreespaceStreamingStatus.Reset)
        self.block_updates = False
        self.lockers.clear()
        self.pausers.clear()
        self.__apply_streaming_settings_and_update_state()
        if self.num_commands_registered != 0:
            raise RuntimeError(f"Failed to reset streaming manager. {self.num_commands_registered} commands still registered.\n {self.registered_commands}")
        if self.num_callbacks_registered != 0:
            raise RuntimeError(f"Failed to reset streaming manager. {self.num_callbacks_registered} callbacks still registered.\n {self.callbacks}")
        return True

    def update(self):
        if self.paused or not self.enabled or not self.sensor.is_streaming: return

        self.apply_updated_settings()
        self.sensor.updateStreaming()
        result = self.sensor.getOldestStreamingPacket()
        if result is not None:
            while result is not None:
                self.sample_count += 1
                self.last_response = result
                slot_index = 0
                for data in result.data:
                    while not self.slots[slot_index].active: slot_index += 1
                    cmd = self.slots[slot_index]
                    info = (cmd.cmd, cmd.param)
                    self.results[info] = data
                    slot_index += 1
                
                #Let all the callbacks know the data was updated
                for cb in self.callbacks.values():
                    if cb.only_newest: continue
                    cb.func(ThreespaceStreamingStatus.Data)

                result = self.sensor.getOldestStreamingPacket()
            
            for cb in self.callbacks.values():
                if cb.only_newest:
                    cb.func(ThreespaceStreamingStatus.Data)
                cb.func(ThreespaceStreamingStatus.DataEnd)

    def register_callback(self, callback: Callable[[ThreespaceStreamingStatus],None], hz=None, only_newest=False):
        if callback in self.callbacks: return
        self.callbacks[callback] = ThreespaceStreamingManager.Callback(callback, hz, only_newest)
        self.__update_streaming_speed()

    def unregister_callback(self, callback: Callable[[ThreespaceStreamingStatus],None]):
        if callback not in self.callbacks: return
        del self.callbacks[callback]
        self.__update_streaming_speed()

    def register_command(self, owner: object, command: StreamableCommands|ThreespaceStreamingOption, param=None, immediate_update=True):
        """
        Adds the given command to the streaming slots and starts streaming it

        Parameters
        ----
        owner : A reference to the object registering the command. A command is only unregistered after all its owners release it
        command : The command to register
        param : The parameter (if any) required for the command to be streamed. The command and param together identify a single slot
        immediate_update : If true, the streaming manager will immediately change the streaming slots on the sensor. If doing bulk registers, it
        is useful to set this as False until the last one for performance purposes.

        Returns
        -------
        True : Successfully registered the command
        False : Failed to register the command. Streaming slots are full
        """
        if isinstance(command, tuple):
            param = command[1]
            command = command[0]
        info = (command, param)
        if info in self.registered_commands: #Already registered, just add this as an owner
            cmd = self.registered_commands[info]
            if len(cmd.registrations) == 0 and self.num_commands_registered >= self.num_slots: #No room to register
                return False
            
            cmd.registrations.add(owner)
            if immediate_update and self.dirty:
                return self.__apply_streaming_settings_and_update_state()
            return True
        
        if self.locked: #Wasn't already registered, so don't allow new registrations
            return False
        
        #Make sure to only consider a command registered if it has registrations
        num_commands_registered = self.num_commands_registered
        if num_commands_registered >= self.num_slots: #No room to register
            return False
        
        #Register the command and add it to the streaming
        self.registered_commands[info] = ThreespaceStreamingManager.Command(command, param=param, slot=num_commands_registered)
        self.registered_commands[info].labels = self.sensor.getStreamingLabel(command.value).data
        self.registered_commands[info].registrations.add(owner)
        self.dirty = True
        if immediate_update:
            return self.__apply_streaming_settings_and_update_state()
        return True

    def unregister_command(self, owner: object, command: StreamableCommands|ThreespaceStreamingOption, param=None, immediate_update=True):
        """
        Removes the given command to the streaming slots and starts streaming it

        Parameters
        ----------
        owner : A reference to the object unregistering the command. A command is only unregistered after all its owners release it
        command : The command to unregister
        param : The param (if any) required for the command
        immediate_update : If true, the streaming manager will immediately change the streaming slots on the sensor. If doing bulk unregisters, it
        is useful to set this as False until the last one for performance purposes.
        """
        if isinstance(command, tuple):
            param = command[1]
            command = command[0]
        info = (command, param)
        if info not in self.registered_commands:
            return
        
        try:
            self.registered_commands[info].registrations.remove(owner)
        except KeyError:
            return #This owner wasn't registered to begin with, just ignore
        
        #Nothing else to do
        if len(self.registered_commands[info].registrations) != 0: 
            return
        
        #Remove the command from streaming since nothing owns it anymore
        self.dirty = True
        if immediate_update:
            self.__apply_streaming_settings_and_update_state()


    def __build_stream_slots_string(self):
        cmd_strings = []
        self.slots.clear()
        if self.num_commands_registered == 0: return "255"
        i = 0
        for cmd_key in self.registered_commands:
            cmd = self.registered_commands[cmd_key]
            if not cmd.active: continue #Skip non active registrations
            self.slots.append(cmd)
            cmd.slot = i
            if cmd.param == None:
                cmd_strings.append(str(cmd.cmd.value))
            else:
                cmd_strings.append(f"{cmd.cmd.value}:{cmd.param}")
            i += 1
        return ','.join(cmd_strings)

    #More user friendly version of __apply_streaming_settings_and_update_state that prevents the user from calling it when not needed.
    def apply_updated_settings(self):
        """
        This applys the current settings of the streaming manager and updates its state. This is normally done automatically, however
        if the user is registering/unregistering with immediate_update turned off, this can be called to force the update.
        """
        if not self.dirty: return self.validated
        return self.__apply_streaming_settings_and_update_state()

    def __apply_streaming_settings_and_update_state(self, ignore_lock=False):
        """
        Used to apply the current configuration this manager represents to the streaming.
        This involves disabling streaming if currently running
        """
        if self.block_updates or (self.locked and not ignore_lock):
            return False
        
        if self.sensor.is_streaming:
            self.__stop_streaming()
        
        #Clean up any registrations that need removed and activate any that need activated
        if self.dirty:
            to_remove = []
            for k, v in self.registered_commands.items():
                if len(v.registrations) == 0:
                    to_remove.append(k)
                    continue
                v.active = True
            for key in to_remove:
                del self.registered_commands[key]
                if key in self.results:
                    del self.results[key]
        self.dirty = False

        if self.num_commands_registered > 0:
            slots_string = self.__build_stream_slots_string()
            err, num_successes = self.sensor.set_settings(stream_slots=slots_string, stream_interval=self.interval)
            if err:
                self.validated = False
                return False
            if not self.paused and self.enabled: 
                self.__start_streaming() #Re-enable
        
        self.validated = True
        return True

    def __update_streaming_speed(self):        
        required_interval = None
        for callback in self.callbacks.values():
            if callback.interval is None: continue
            if required_interval is None or callback.interval < required_interval:
                required_interval = callback.interval

        if required_interval is None: #Treat required as current to make sure the current interval is still valid
            required_interval = self.interval 

        required_interval = min(self.max_interval, max(self.min_interval, required_interval))
        if required_interval != self.interval:
            print(f"Updating streaming speed from {1000000 / self.interval}hz to {1000000 / required_interval}hz")
            self.interval = int(required_interval)
            self.dirty = True
            self.__apply_streaming_settings_and_update_state()

    def __start_streaming(self):
        self.sensor.startStreaming()
        self.is_streaming = True

    def __stop_streaming(self):
        self.sensor.stopStreaming()
        self.is_streaming = False

    @property
    def num_commands_registered(self):
        return len([v for v in self.registered_commands.values() if len(v.registrations) != 0])
    
    @property
    def num_callbacks_registered(self):
        return len(self.callbacks)

    def get_value(self, command: StreamableCommands|ThreespaceStreamingOption, param=None):
        if isinstance(command, tuple):
            param = command[1]
            command = command[0]
        return self.results.get((command, param), None)

    def get_last_response(self):
        return self.last_response

    def get_header(self):
        return self.last_response.header  

    def get_cmd_labels(self):
        return ','.join(cmd.labels for cmd in self.registered_commands.values())
    
    def get_header_labels(self):
        order = threespaceGetHeaderLabels(self.sensor.header_info)
        return ','.join(order)
    
    def get_response_labels(self):
        return ','.join([self.get_header_labels(), self.get_cmd_labels()])

    def enable(self):
        if self.enabled:
            return
        self.enabled = True
        self.__apply_streaming_settings_and_update_state()

    def disable(self):
        if not self.enabled:
            return
        if self.is_streaming:
            self.__stop_streaming()
        self.enabled = False

    def set_max_hz(self, hz: float):
        if hz <= 0 or hz > 2000: 
            raise ValueError(f"Invalid streaming Hz {hz}")
        self.min_interval = 1000000 // hz
        self.__update_streaming_speed()

    def set_min_hz(self, hz: float):
        if hz <= 0 or hz > 2000: 
            raise ValueError(f"Invalid streaming Hz {hz}")
        self.max_interval = 1000000 // hz
        self.__update_streaming_speed()

    def get_slots_from_sensor(self):
        """
        get a list containing the streaming information from the current sensor
        """
        slot_setting: str = self.sensor.get_settings("stream_slots")
        slots = slot_setting.split(',')
        slot_info = []
        for slot in slots:
            info = slot.split(':')
            slot = int(info[0]) #Ignore parameters if any
            param = None
            if len(info) > 1:
                param = int(info[1])
            if slot != 255:
                slot_info.append((slot, param))
            else:
                slot_info.append(None)
        
        return slot_info


from yostlabs.tss3.consts import *
from yostlabs.communication.base import ThreespaceInputStream, ThreespaceOutputStream, ThreespaceComClass
from yostlabs.communication.serial import ThreespaceSerialComClass

from enum import Enum
from dataclasses import dataclass, field
from typing import TypeVar, Generic
import struct
import types
import inspect
import time
import math


#For converting from internal format specifiers to struct module specifiers
__3space_format_conversion_dictionary = {
    'f': {"c": 'f', "size": 4},
    'd' : {"c": 'd', "size": 8},

    'b' : {"c": 'B', "size": 1},
    'B' : {"c": 'H', "size": 2},
    "u" : {"c": 'L', "size": 4},
    "U" : {"c": 'Q', "size": 8},

    "i" : {"c": 'b', "size": 1},
    "I" : {"c": 'h', "size": 2},
    "l" : {"c": 'l', "size": 4},
    "L" : {"c": 'q', "size": 8},

    #Strings actually don't convert, they need handled special because
    #struct unpack assumes static length strings, whereas the sensors
    #use variable length null terminated strings
    "s" : {"c": 's', "size": float('nan')},
    "S" : {"c": 's', "size": float('nan')}
}

def _3space_format_get_size(format_str: str):
    size = 0
    for c in format_str:
        size += __3space_format_conversion_dictionary[c]["size"]
    return size

def _3space_format_to_external(format_str: str):
    return ''.join(__3space_format_conversion_dictionary[c]['c'] for c in format_str)

@dataclass
class ThreespaceCommandInfo:
    name: str
    num: int
    in_format: str
    out_format: str

    num_out_params: int = field(init=False)
    out_size: int = field(init=False,)

    def __post_init__(self):
        self.num_out_params = len(self.out_format)
        self.out_size = _3space_format_get_size(self.out_format)

class ThreespaceCommand:

    BINARY_START_BYTE = 0xf7
    BINARY_START_BYTE_HEADER = 0xf9

    def __init__(self, name: str, num: int, in_format: str, out_format: str):
        self.info = ThreespaceCommandInfo(name, num, in_format, out_format)
        self.in_format = _3space_format_to_external(self.info.in_format)
        self.out_format = _3space_format_to_external(self.info.out_format)

    def format_cmd(self, *args, header_enabled=False):
        cmd_data = struct.pack("<B", self.info.num)
        for i, c in enumerate(self.in_format):
            if c != 's':
                cmd_data += struct.pack(f"<{c}", args[i])
            else:
                cmd_data += struct.pack(f"<{len(args[i])}sb", bytes(args[i], 'ascii'), 0)
        checksum = sum(cmd_data) % 256
        start_byte = ThreespaceCommand.BINARY_START_BYTE_HEADER if header_enabled else ThreespaceCommand.BINARY_START_BYTE
        return struct.pack(f"<B{len(cmd_data)}sB", start_byte, cmd_data, checksum)

    def send_command(self, com: ThreespaceOutputStream, *args, header_enabled = False):
        cmd = self.format_cmd(*args, header_enabled=header_enabled)
        com.write(cmd)

    #Read the command result from an already read buffer. This will modify the given buffer to remove
    #that data as well
    def parse_response(self, response: bytes):
        if self.info.num_out_params == 0: return None
        output = []
        
        if math.isnan(self.info.out_size): #Has strings in it, must slow parse
            for c in self.out_format:
                if c != 's':
                    format_str = f"<{c}"
                    size = struct.calcsize(format_str)
                    output.append(struct.unpack(format_str, response[:size])[0])
                    #TODO: Switch to using numpy views instead of slicing
                    response = response[size:]
                else: #Strings are special, find the null terminator
                    str_len = response.index(0)
                    output.append(struct.unpack(f"<{str_len}s", response[str_len])[0])
                    response = response[str_len + 1:] #+1 to skip past the null terminator character too
        else: #Fast parse because no strings
            output.extend(struct.unpack(f"<{self.out_format}", response[:self.info.out_size]))

        
        if self.info.num_out_params == 1:
            return output[0]
        return output

    #Read the command dynamically from an input stream
    def read_command(self, com: ThreespaceInputStream):
        raw = bytearray([])
        if self.info.num_out_params == 0: return None, raw
        output = []
        for c in self.out_format:
            if c != 's':
                format_str = f"<{c}"
                size = struct.calcsize(format_str)
                response = com.read(size)
                raw += response
                if len(response) != size:
                    print(f"Failed to read {c} type. Aborting...")
                    return None
                output.append(struct.unpack(format_str, response)[0])
            else: #Strings are special, find the null terminator
                response = com.read(1)
                raw += response
                if len(response) != 1:
                    print(f"Failed to read string. Aborting...")
                    return None
                byte = chr(response[0])
                string = ""
                while byte != '\0':
                    string += byte
                    #Get next byte
                    response = com.read(1)
                    raw += response
                    if len(response) != 1:
                        print(f"Failed to read string. Aborting...")
                        return None
                    byte = chr(response[0])
                output.append(string)
        
        if self.info.num_out_params == 1:
            return output[0], raw
        return output, raw

class ThreespaceGetStreamingBatchCommand(ThreespaceCommand):

    def __init__(self, streaming_slots: list[ThreespaceCommand]):
        self.commands = streaming_slots
        combined_out_format = ''.join(slot.info.out_format for slot in streaming_slots if slot is not None)
        super().__init__("getStreamingBatch", THREESPACE_GET_STREAMING_BATCH_COMMAND_NUM, "", combined_out_format)
        self.out_format = ''.join(slot.out_format for slot in streaming_slots if slot is not None)

    def set_stream_slots(self, streaming_slots: list[ThreespaceCommand]):
        self.commands = streaming_slots
        self.out_format = ''.join(slot.out_format for slot in streaming_slots if slot is not None)

    def parse_response(self, response: bytes):
        data = []
        for command in self.commands:
            if command is None: continue
            cmd_response_size = command.info.out_size
            data.append(command.parse_response(response))
            response = response[cmd_response_size:]
        
        return data
    
    def read_command(self, com: ThreespaceInputStream):
        #Get the response to all the streaming commands
        response = []
        raw_response = bytearray([])
        for command in self.commands:
            if command is None: continue
            binary = com.read(command.info.out_size)
            raw_response += binary
            out = command.parse_response(binary)
            response.append(out)
        
        return response, raw_response

THREESPACE_HEADER_FORMAT_CHARS = ['b', 'L', 'B', 'B', 'L', 'H']

@dataclass
class ThreespaceHeaderInfo:
    __bitfield: int = 0
    format: str = ""
    size: int = 0

    def get_start_byte(self, header_field: int):
        """
        Given a header field, give the initial byte offset for that field when
        using binary mode
        """
        if not header_field & self.__bitfield: return None #The bit is not enabled, no start byte
        #Get the index of the bit
        bit_pos = 0
        header_field >>= 1
        while header_field > 0:
            bit_pos += 1
            header_field >>= 1

        #Add up the size of everything before this field
        start = 0
        for i in range(bit_pos):
            if (1 << i) & self.__bitfield:
                start += struct.calcsize(THREESPACE_HEADER_FORMAT_CHARS[i])
        return start
    
    def get_index(self, header_field: int):
        if not header_field & self.__bitfield: return None
        index = 0
        bit = 1
        while bit < header_field:
            if bit & self.__bitfield:
                index += 1
            bit <<= 1
        return index

    def __update(self):
        self.format = "<"
        for i in range(THREESPACE_HEADER_NUM_BITS):
            if self.__bitfield & (1 << i):
                self.format += THREESPACE_HEADER_FORMAT_CHARS[i]
        self.size = struct.calcsize(self.format)

    @property
    def bitfield(self):
        return self.__bitfield
    
    @bitfield.setter
    def bitfield(self, value):
        self.__bitfield = value
        self.__update()
    
    @property
    def status_enabled(self):
        return bool(self.__bitfield & THREESPACE_HEADER_STATUS_BIT)
    
    @status_enabled.setter
    def status_enabled(self, value: bool):
        if value: self.__bitfield |= THREESPACE_HEADER_STATUS_BIT
        else: self.__bitfield &= ~THREESPACE_HEADER_STATUS_BIT
        self.__update()
    
    @property
    def timestamp_enabled(self):
        return bool(self.__bitfield & THREESPACE_HEADER_TIMESTAMP_BIT)
    
    @timestamp_enabled.setter
    def timestamp_enabled(self, value: bool):
        if value: self.__bitfield |= THREESPACE_HEADER_TIMESTAMP_BIT
        else: self.__bitfield &= ~THREESPACE_HEADER_TIMESTAMP_BIT
        self.__update()

    @property
    def echo_enabled(self):
        return bool(self.__bitfield & THREESPACE_HEADER_ECHO_BIT)
    
    @echo_enabled.setter
    def echo_enabled(self, value: bool):
        if value: self.__bitfield |= THREESPACE_HEADER_ECHO_BIT
        else: self.__bitfield &= ~THREESPACE_HEADER_ECHO_BIT
        self.__update()       

    @property
    def checksum_enabled(self):
        return bool(self.__bitfield & THREESPACE_HEADER_CHECKSUM_BIT)
    
    @checksum_enabled.setter
    def checksum_enabled(self, value: bool):
        if value: self.__bitfield |= THREESPACE_HEADER_CHECKSUM_BIT
        else: self.__bitfield &= ~THREESPACE_HEADER_CHECKSUM_BIT     
        self.__update()

    @property
    def serial_enabled(self):
        return bool(self.__bitfield & THREESPACE_HEADER_SERIAL_BIT)
    
    @serial_enabled.setter
    def serial_enabled(self, value: bool):
        if value: self.__bitfield |= THREESPACE_HEADER_SERIAL_BIT
        else: self.__bitfield &= ~THREESPACE_HEADER_SERIAL_BIT  
        self.__update()

    @property
    def length_enabled(self):
        return bool(self.__bitfield & THREESPACE_HEADER_LENGTH_BIT)
    
    @length_enabled.setter
    def length_enabled(self, value: bool):
        if value: self.__bitfield |= THREESPACE_HEADER_LENGTH_BIT
        else: self.__bitfield &= ~THREESPACE_HEADER_LENGTH_BIT      
        self.__update()              


@dataclass
class ThreespaceHeader:
    raw: tuple = field(default=None, repr=False)

    #Order here matters
    status: int = None
    timestamp: int = None
    echo: int = None
    checksum: int = None
    serial: int = None
    length: int = None

    raw_binary: bytes = field(repr=False, default_factory=lambda: bytes([]))
    info: ThreespaceHeaderInfo = field(default_factory=lambda: ThreespaceHeaderInfo(), repr=False)

    @staticmethod
    def from_tuple(data, info: ThreespaceHeaderInfo):
        raw_expanded = []
        cur_index = 0
        for i in range(THREESPACE_HEADER_NUM_BITS):
            if info.bitfield & (1 << i): 
                raw_expanded.append(data[cur_index])
                cur_index += 1
            else:
                raw_expanded.append(None)
        return ThreespaceHeader(data, *raw_expanded, info=info)

    @staticmethod
    def from_bytes(byte_data: bytes, info: ThreespaceHeaderInfo):
        if info.size == 0: return ThreespaceHeader()
        header = ThreespaceHeader.from_tuple(struct.unpack(info.format, byte_data[:info.size]), info)
        header.raw_binary = byte_data
        return header

    def __getitem__(self, key):
        return self.raw[key]
    
    def __len__(self):
        return len(self.raw)
    
    def __iter__(self):
        return iter(self.raw)

class StreamableCommands(Enum):
    GetTaredOrientation = 0
    GetTaredOrientationAsEuler = 1
    GetTaredOrientationAsMatrix = 2
    GetTaredOrientationAsAxisAngle = 3
    GetTaredOrientationAsTwoVector = 4

    GetDifferenceQuaternion = 5

    GetUntaredOrientation = 6
    GetUntaredOrientationAsEuler = 7
    GetUntaredOrientationAsMatrix = 8
    GetUntaredOrientationAsAxisAngle = 9
    GetUntaredOrientationAsTwoVector = 10

    GetTaredOrientationAsTwoVectorSensorFrame = 11
    GetUntaredOrientationAsTwoVectorSensorFrame = 12

    GetPrimaryBarometerPressure = 13
    GetPrimaryBarometerAltitude = 14
    GetBarometerAltitudeById = 15
    GetBarometerPressureById = 16

    GetAllPrimaryNormalizedData = 32
    GetPrimaryNormalizedGyroRate = 33
    GetPrimaryNormalizedAccelVec = 34
    GetPrimaryNormalizedMagVec = 35
    
    GetAllPrimaryCorrectedData = 37
    GetPrimaryCorrectedGyroRate = 38
    GetPrimaryCorrectedAccelVec = 39
    GetPrimaryCorrectedMagVec = 40

    GetPrimaryGlobalLinearAccel = 41
    GetPrimaryLocalLinearAccel = 42

    GetTemperatureCelsius = 43
    GetTemperatureFahrenheit = 44
    GetMotionlessConfidenceFactor = 45

    GetNormalizedGyroRate = 51
    GetNormalizedAccelVec = 52
    GetNormalizedMagVec = 53

    GetCorrectedGyroRate = 54
    GetCorrectedAccelVec = 55
    GetCorrectedMagVec = 56

    GetRawGyroRate = 65
    GetRawAccelVec = 66
    GetRawMagVec = 67

    GetEeptsOldestStep = 70
    GetEeptsNewestStep = 71
    GetEeptsNumStepsAvailable = 72

    GetTimestamp = 94

    GetBatteryVoltage = 201
    GetBatteryPercent = 202
    GetBatteryStatus = 203

    GetGpsCoord = 215
    GetGpsAltitude = 216
    GetGpsFixState = 217
    GetGpsHdop = 218
    GetGpsSattelites = 219

    GetButtonState = 250

THREESPACE_AWAIT_COMMAND_FOUND = 0
THREESPACE_AWAIT_COMMAND_TIMEOUT = 1

T = TypeVar('T')

@dataclass
class ThreespaceCmdResult(Generic[T]):
    raw: tuple = field(default=None, repr=False)

    header: ThreespaceHeader = None
    data: T = None
    raw_data: bytes = field(default=None, repr=False)

    def __init__(self, data: T, header: ThreespaceHeader, data_raw_binary: bytes = None):
        self.header = header
        self.data = data
        self.raw = (header.raw, data)
        self.raw_data = data_raw_binary

    def __getitem__(self, key):
        return self.raw[key]
    
    def __len__(self):
        return len(self.raw)
    
    def __iter__(self):
        return iter(self.raw)   
    
    @property
    def raw_binary(self):
        bin = bytearray([])
        if self.header is not None and self.header.raw_binary is not None:
            bin += self.header.raw_binary
        if self.raw_data is not None:
            bin += self.raw_data
        return bin

@dataclass
class ThreespaceBootloaderInfo:
    memstart: int
    memend: int
    pagesize: int
    bootversion: int

#Required for the API to work. The API will attempt to keep these enabled at all times.
THREESPACE_REQUIRED_HEADER = THREESPACE_HEADER_ECHO_BIT | THREESPACE_HEADER_CHECKSUM_BIT | THREESPACE_HEADER_LENGTH_BIT
class ThreespaceSensor:
    
    def __init__(self, com = None, timeout=2):
        if com is None: #Default to attempting to use the serial com class if none is provided
            com = ThreespaceSerialComClass
        
        #Auto discover using the supplied com class type
        if inspect.isclass(com) and issubclass(com, ThreespaceComClass):
            new_com = None
            for serial_com in com.auto_detect():
                new_com = serial_com
                break #Exit after getting 1
            if new_com is None:
                raise RuntimeError("Failed to auto discover com port")
            self.com = new_com
            self.com.open()
        #The supplied com already was a com class, nothing to do
        elif inspect.isclass(type(com)) and issubclass(type(com), ThreespaceComClass):
            self.com = com
        else: #Unknown type, try making a ThreespaceSerialComClass out of this
            try:
                self.com = ThreespaceSerialComClass(com)
            except:
                raise ValueError("Failed to create default ThreespaceSerialComClass from parameter:", type(com), com)

        self.com.read_all() #Clear anything that may be there

        self.commands: list[ThreespaceCommand] = [None] * 256
        self.getStreamingBatchCommand: ThreespaceGetStreamingBatchCommand = None
        self.funcs = {}
        for command in _threespace_commands:
            #Some commands are special and need added specially
            if command.info.num == THREESPACE_GET_STREAMING_BATCH_COMMAND_NUM:
                self.getStreamingBatchCommand = ThreespaceGetStreamingBatchCommand([])
                command = self.getStreamingBatchCommand
            
            self.__add_command(command)

        self.immediate_debug = False
        self.misaligned = False
        self.dirty_cache = False
        self.header_enabled = True 

        #All the different streaming options
        self.is_data_streaming = False
        self.is_log_streaming = False
        self.is_file_streaming = False
        self._force_stop_streaming()

        #Used to ensure connecting to the correct sensor when reconnecting
        self.serial_number = None

        self.__cached_in_bootloader = self.__check_bootloader_status()
        if not self.in_bootloader:
            self.__firmware_init()
        else:
            self.serial_number = self.bootloader_get_sn()

    def __firmware_init(self):
        """
        Should only be called when not streaming and known in firmware.
        Called for powerup events when booting into firmware
        """
        self.dirty_cache = False #No longer dirty cause initializing

        self.com.read_all() #Clear anything that may be there
        
        self.__reinit_firmware()
        
        self.valid_mags = self.__get_valid_components("valid_mags")
        self.valid_accels = self.__get_valid_components("valid_accels")
        self.valid_gyros = self.__get_valid_components("valid_gyros")
        self.valid_baros = self.__get_valid_components("valid_baros")

    def __get_valid_components(self, key: str):
        valid = self.get_settings(key)
        if len(valid) == 0: return []
        return [int(v) for v in valid.split(',')]

    def __reinit_firmware(self):
        """
        Called when settings may have changed but a full reboot did not occur
        """
        self.com.read_all() #Clear anything that may be there
        self.dirty_cache = False #No longer dirty cause initializing
        
        self.header_info = ThreespaceHeaderInfo()
        self.cmd_echo_byte_index = None
        self.streaming_slots: list[ThreespaceCommand] = [None] * 16
        self.streaming_packets: list[ThreespaceCmdResult[list]] = []

        self.file_stream_data = bytearray([])
        self.file_stream_length = 0

        self.streaming_packet_size = 0
        self.header_enabled = True
        self._force_stop_streaming()

        #Now reinitialize the cached settings
        self.__cache_header_settings()
        self.cache_streaming_settings()

        self.serial_number = int(self.get_settings("serial_number"), 16)
        self.immediate_debug = int(self.get_settings("debug_mode")) == 1 #Needed for some startup processes when restarting

    def __add_command(self, command: ThreespaceCommand):
        if self.commands[command.info.num] != None:
            print(f"Registering duplicate command: {command.info.num} {self.commands[command.info.num].info.name} {command.info.name}")
        self.commands[command.info.num] = command

        #Build the actual method for executing the command
        code = f"def {command.info.name}(self, *args):\n"
        code += f"    return self.execute_command(self.commands[{command.info.num}], *args)"
        exec(code, globals(), self.funcs)
        setattr(self, command.info.name, types.MethodType(self.funcs[command.info.name], self))

    def __get_command(self, command_name: str):
        for command in self.commands:
            if command is None: continue
            if command.info.name == command_name:
                return command
        return None

    @property
    def is_streaming(self):
        return self.is_data_streaming or self.is_log_streaming or self.is_file_streaming

    #Can't just do if "header" in string because log_header_enabled exists and doesn't actually require cacheing the header
    HEADER_KEYS = ["header", "header_status", "header_timestamp", "header_echo", "header_checksum", "header_serial", "header_length"]
    def set_settings(self, param_string: str = None, **kwargs):
        self.check_dirty()
        #Build cmd string
        params = []
        if param_string is not None:
            params.append(param_string)
        
        for key, value in kwargs.items():
            if isinstance(value, list):
                value = [str(v) for v in value]
                value = ','.join(value)
            elif isinstance(value, bool):
                value = int(value)
            params.append(f"{key}={value}")
        cmd = f"!{';'.join(params)}\n"

        #For dirty check
        keys = cmd[1:-1].split(';')
        keys = [v.split('=')[0] for v in keys]

        #Send cmd
        self.com.write(cmd.encode())

        #Default values
        err = 3
        num_successes = 0

        #Read response
        if self.is_streaming: #Streaming have to read via peek and also validate it more
            max_response_length = len("255,255\r\n")
            found_response = False
            start_time = time.time()
            while not found_response: #Infinite loop to wait for the data to be available
                if time.time() - start_time > self.com.timeout:
                    print("Timed out waiting for set_settings response")
                    return err, num_successes
                line = ""
                while True: #A loop used to allow breaking out of to be less wet.
                    line = self.com.peekline(max_length=max_response_length)
                    if b'\n' not in line:
                        break
                    
                    try:
                        values = line.decode().strip()
                        values = values.split(',')
                        if len(values) != 2: break
                        err = int(values[0])
                        num_successes = int(values[1])
                    except: break
                    if err > 255 or num_successes > 255:
                        break

                    #Successfully got pass all the checks!
                    #Consume the buffer and continue
                    found_response = True
                    self.com.readline()
                    break
                if found_response: break
                while not self.updateStreaming(max_checks=1): pass #Wait for streaming to parse something!
        else:
            #When not streaming, way more straight forward
            try:
                response = self.com.readline()
                response = response.decode().strip()
                err, num_successes = response.split(',')
                err = int(err)
                num_successes = int(num_successes)    
            except:
                print("Failed to parse set response:", response)
                return err, num_successes
        
        #Handle updating state variables based on settings
        #If the user modified the header, need to cache the settings so the API knows how to interpret responses
        if "header" in cmd.lower(): #First do a quick check
            if any(v in keys for v in ThreespaceSensor.HEADER_KEYS): #Then do a longer check
                self.__cache_header_settings()
        
        if "stream_slots" in cmd.lower():
            self.cache_streaming_settings()
        
        if any(v in keys for v in ("default", "reboot")): #All the settings changed, just need to mark dirty
            self.set_cached_settings_dirty()

        if err:
            print(f"Err setting {cmd}: {err=} {num_successes=}")
        return err, num_successes

    def get_settings(self, *args: str) -> dict[str, str] | str:
        self.check_dirty()
        #Build and send the cmd
        params = list(args)
        cmd = f"?{';'.join(params)}\n"
        self.com.write(cmd.encode())

        keys = cmd[1:-1].split(';')
        error_response = "<KEY_ERROR>"

        #Wait for the response to be available if streaming
        #NOTE: THIS WILL NOT WORK WITH SETTINGS SUCH AS ?all ?settings or QUERY STRINGS
        #THIS can be worked around by first getting a setting that does echo normally, as that will allow
        #the sensor to determine where the ascii data actually starts.
        #Ex: get_settings("header", "all") would work
        if self.is_streaming:
            first_key = bytes(keys[0] + "=", 'ascii') #Add on the equals sign to try and make this less likely to conflict with binary data
            possible_outputs = [(len(error_response), bytes(error_response, 'ascii')), (len(first_key), first_key)]
            possible_outputs.sort() #Must try the smallest one first because if streaming is slow, may take a while for the data to fill pass the largest possible value
            start_time = time.time()
            while True:
                if time.time() - start_time > self.com.timeout:
                    print("Timeout parsing get response")
                    return {}
                found_response = False
                for length, key in possible_outputs:
                    possible_response = self.com.peek(length)
                    if possible_response == key: #This the response, so break and parse
                        found_response = True
                        break
                if found_response: break
                while not self.updateStreaming(max_checks=1): pass #Wait for streaming to process something. May just advance due to invalid
        
        #Read the response
        try:
            response = self.com.readline()
            if ord('\n') not in response:
                print("Failed to get whole line")
            response = response.decode().strip().split(';')
        except:
            print("Failed to parse get:", response)
        
        #Build the response dict
        response_dict = {}
        for i, v in enumerate(response):
            if v == error_response:
                response_dict[keys[i]] = error_response
                continue
            try:
                key, value = v.split('=')
                response_dict[key] = value
            except:
                print("Failed to parse get:", response)
        
        #Format response
        if len(response_dict) == 1:
            return list(response_dict.values())[0]
        return response_dict

    def execute_command(self, cmd: ThreespaceCommand, *args):
        self.check_dirty()

        retries = 0
        MAX_RETRIES = 3

        while retries < MAX_RETRIES:
            cmd.send_command(self.com, *args, header_enabled=self.header_enabled)
            result = self.__await_command(cmd)
            if result == THREESPACE_AWAIT_COMMAND_FOUND:
                break
            retries += 1
        
        if retries == MAX_RETRIES:
            raise RuntimeError(f"Failed to get response to command {cmd.info.name}")

        return self.read_and_parse_command(cmd)
    
    def read_and_parse_command(self, cmd: ThreespaceCommand):
        if self.header_enabled:
            header = ThreespaceHeader.from_bytes(self.com.read(self.header_info.size), self.header_info)
        else:
            header = ThreespaceHeader()
        result, raw = cmd.read_command(self.com)
        return ThreespaceCmdResult(result, header, data_raw_binary=raw)

    def __peek_checksum(self, header: ThreespaceHeader):
        header_len = len(header.raw_binary)
        data = self.com.peek(header_len + header.length)[header_len:]
        if len(data) != header.length: return False
        checksum = sum(data) % 256
        return checksum == header.checksum

    def __await_command(self, cmd: ThreespaceCommand, timeout=2):
        start_time = time.time()

        #Update the streaming until the result for this command is next in the buffer
        while True:
            if time.time() - start_time > timeout:
                return THREESPACE_AWAIT_COMMAND_TIMEOUT
            
            #Get potential header
            header = self.com.peek(self.header_info.size)
            if len(header) != self.header_info.size: #Wait for more data
                continue

            #Check to see what this packet is a response to
            header = ThreespaceHeader.from_bytes(header, self.header_info)
            echo = header.echo

            if echo == cmd.info.num: #Cmd matches
                if self.__peek_checksum(header):
                    return THREESPACE_AWAIT_COMMAND_FOUND
                
                #Error in packet, go start realigning
                if not self.misaligned:
                    print(f"Checksum mismatch for command {cmd.info.num}")
                    self.misaligned = True
                self.com.read(1)
            else:
                #It wasn't a response to the command, so may be a response to some internal system
                self.__internal_update(header)

    def __internal_update(self, header: ThreespaceHeader):
        """
        This should be called after a header is obtained via a command and it is determined that it can't
        be in response to a synchronous command that got sent. This manages updating the streaming and realigning
        the data buffer
        """
        checksum_match = False #Just for debugging

        #NOTE: FOR THIS TO WORK IT IS REQUIRED THAT THE HEADER DOES NOT CHANGE WHILE STREAMING ANY FORM OF DATA.
        #IT IS UP TO THE API TO ENFORCE NOT ALLOWING HEADER CHANGES WHILE ANY OF THOSE THINGS ARE HAPPENING
        if self.is_data_streaming and header.echo == THREESPACE_GET_STREAMING_BATCH_COMMAND_NUM: #Its a streaming packet, so update streaming
            if checksum_match := self.__peek_checksum(header):
                self.__update_base_streaming()
                return True
        elif self.is_log_streaming and header.echo == THREESPACE_FILE_READ_BYTES_COMMAND_NUM:
            if checksum_match := self.__peek_checksum(header):
                self.__update_log_streaming()
                return True
        elif self.is_file_streaming and header.echo == THREESPACE_FILE_READ_BYTES_COMMAND_NUM:
            if checksum_match := self.__peek_checksum(header):
                self.__update_file_streaming()
                return True

        #The response didn't match any of the expected asynchronous streaming API responses, so assume a misalignment
        #and start reading through the buffer
        if not self.misaligned:
            print(f"Possible Misalignment or corruption/debug message, header {header} raw {[hex(v) for v in header.raw_binary]}, Checksum match? {checksum_match}")
            self.misaligned = True
        self.com.read(1) #Because of expected misalignment, go through buffer 1 by 1 until realigned

    def updateStreaming(self, max_checks=float('inf')):
        """
        Returns true if any amount of data was processed whether valid or not
        """
        if not self.is_streaming: return False

        #I may need to make this have a max num bytes it will process before exiting to prevent locking up on slower machines
        #due to streaming faster then the program runs
        num_checks = 0
        data_processed = False
        while num_checks < max_checks:
            if self.com.length < self.header_info.size:
                return data_processed
            
            #Get header
            header = self.com.peek(self.header_info.size)

            #Get the header and send it to the internal update
            header = ThreespaceHeader.from_bytes(header, self.header_info)
            self.__internal_update(header)
            data_processed = True #Internal update always processes data. Either reads a streaming message, or advances buffer due to misalignment
            num_checks += 1
        
        return data_processed


    def startStreaming(self):
        if self.is_data_streaming: return
        self.check_dirty()
        self.streaming_packets.clear()

        self.header_enabled = True

        self.cache_streaming_settings()

        cmd = self.commands[85]
        cmd.send_command(self.com, header_enabled=self.header_enabled)
        if self.header_enabled:
            self.__await_command(cmd)
            header = ThreespaceHeader.from_bytes(self.com.read(self.header_info.size), self.header_info)
        else:
            header = ThreespaceHeader()
        self.is_data_streaming = True
        return ThreespaceCmdResult(None, header)

    def _force_stop_streaming(self):
        """
        This function is used to stop streaming without validating it was streaming and ignoring any output of the
        communication line. This is a destructive call that will lose data, but will gurantee stopping streaming
        and leave the communication line in a clean state
        """
        cached_header_enabled = self.header_enabled
        cahched_dirty = self.dirty_cache

        #Must set these to gurantee it doesn't try and parse a response from anything
        self.dirty_cache = False
        self.header_enabled = False #Keep off for the attempt at stop streaming since if in an invalid state, won't be able to get response
        self.stopStreaming() #Just in case was streaming
        self.fileStopStream()

        #TODO: Change this to pause the data logging instead, then check the state and update
        self.stopDataLogging()
        
        #Restore
        self.header_enabled = cached_header_enabled
        self.dirty_cache = cahched_dirty

    def stopStreaming(self):
        self.check_dirty()
        cmd = self.commands[86]
        cmd.send_command(self.com, header_enabled=self.header_enabled)
        if self.header_enabled: #Header will be enabled while streaming, but this is useful for startup
            self.__await_command(cmd)
            header = ThreespaceHeader.from_bytes(self.com.read(self.header_info.size), self.header_info)
        else:
            header = ThreespaceHeader()
        time.sleep(0.05)
        while self.com.length:
            self.com.read_all()
        self.is_data_streaming = False
        return ThreespaceCmdResult(None, header)

    def set_cached_settings_dirty(self):
        """
        Could be streaming settings, header settings...
        Basically the sensor needs reinitialized
        """
        self.dirty_cache = True

    def __attempt_rediscover_self(self):
        """
        Trys to change the com class currently being used to be a detected
        com class with the same serial number. Useful for re-enumeration, such as when
        entering bootloader and using USB
        """
        for potential_com in self.com.auto_detect():
            potential_com.open()
            sensor = ThreespaceSensor(potential_com)
            if sensor.serial_number == self.serial_number:
                self.com = potential_com
                return True
            sensor.cleanup() #Handles closing the potential_com
        return False

    def check_dirty(self):
        if not self.dirty_cache: return
        if self.com.reenumerates and not self.com.check_open(): #Must check this, as could have transitioned from bootloader to firmware or vice versa and just needs re-opened/detected
            success = self.__attempt_rediscover_self()
            if not success:
                raise RuntimeError("Sensor connection lost")
        
        self._force_stop_streaming() #Can't be streaming when checking the dirty cache. If you want to stream, don't do things that cause the object to go dirty.
        was_in_bootloader = self.__cached_in_bootloader
        self.__cached_in_bootloader = self.__check_bootloader_status()
        
        if was_in_bootloader and not self.__cached_in_bootloader: #Just Exited bootloader, need to fully reinit
            self.__firmware_init()
        elif not self.__cached_in_bootloader:   #Was already in firmware, so only need to partially reinit
            self.__reinit_firmware()    #Partially init when just naturally dirty
        self.dirty_cache = False

    def cache_streaming_settings(self):
        cached_slots: list[ThreespaceCommand] = []
        slots: str = self.get_settings("stream_slots")
        slots = slots.split(',')
        for slot in slots:
            slot = int(slot.split(':')[0]) #Ignore parameters if any
            if slot != 255:
                cached_slots.append(self.commands[slot])
            else:
                cached_slots.append(None)
        self.streaming_slots = cached_slots.copy()
        self.getStreamingBatchCommand.set_stream_slots(self.streaming_slots)
        self.streaming_packet_size = 0
        for command in self.streaming_slots:
            if command == None: continue
            self.streaming_packet_size += command.info.out_size

    def __cache_header_settings(self):
        """
        Should be called any time changes are made to the header. Will normally be called via the check_dirty/reinit
        """
        header = int(self.get_settings("header"))
        #API requires these bits to be enabled, so don't let them be disabled
        required_header = header | THREESPACE_REQUIRED_HEADER
        if header == self.header_info.bitfield and header == required_header: return #Nothing to update
        
        #Don't allow the header to change while streaming
        #This is to prevent a situation where the header for streaming and commands are different
        #since streaming caches the header. This would cause an issue where the echo byte could be in seperate
        #positions, causing a situation where parsing a command and streaming at the same time breaks since it thinks both are valid cmd echoes.
        if self.is_streaming:
            print("PREVENTING HEADER CHANGE DUE TO CURRENTLY STREAMING")
            self.set_settings(header=self.header_info.bitfield)
            return
        
        if required_header != header:
            print(f"Forcing header checksum, echo, and length enabled")
            self.set_settings(header=required_header)
            return
        
        #Current/New header is valid, so can cache it
        self.header_info.bitfield = header
        self.cmd_echo_byte_index = self.header_info.get_start_byte(THREESPACE_HEADER_ECHO_BIT) #Needed for cmd validation while streaming

    def __update_base_streaming(self):
        """
        Should be called after the packet is validated
        """
        self.streaming_packets.append(self.read_and_parse_command(self.getStreamingBatchCommand))

    def getOldestStreamingPacket(self):
        if len(self.streaming_packets) == 0:
            return None
        return self.streaming_packets.pop(0)
    
    def getNewestStreamingPacket(self):
        if len(self.streaming_packets) == 0:
            return None
        return self.streaming_packets.pop()   
    
    def clearStreamingPackets(self):
        self.streaming_packets.clear()
    
    def fileStartStream(self) -> ThreespaceCmdResult[int]:
        self.check_dirty()
        self.header_enabled = True

        cmd = self.__get_command("__fileStartStream")
        cmd.send_command(self.com, header_enabled=self.header_enabled)
        self.__await_command(cmd)
        
        if self.header_enabled:
            header = ThreespaceHeader.from_bytes(self.com.read(self.header_info.size), self.header_info)
        else:
            header = ThreespaceHeader()

        result, raw = cmd.read_command(self.com)
        self.file_stream_length = result
        self.is_file_streaming = True  

        return ThreespaceCmdResult(result, header, data_raw_binary=raw)
    
    def fileStopStream(self) -> ThreespaceCmdResult[None]:
        self.check_dirty()

        cmd = self.__get_command("__fileStopStream")
        cmd.send_command(self.com, header_enabled=self.header_enabled)

        if self.header_enabled: #Header will be enabled while streaming, but this is useful for startup
            self.__await_command(cmd)
            header = ThreespaceHeader.from_bytes(self.com.read(self.header_info.size), self.header_info)
        else:
            header = ThreespaceHeader()
        
        #TODO: Remove me now that realignment exists and multiple things can be streaming at once
        time.sleep(0.05)
        while self.com.length:
            self.com.read_all()

        self.is_file_streaming = False
        return ThreespaceCmdResult(None, header)

    def getFileStreamData(self):
        to_return = self.file_stream_data.copy()
        self.file_stream_data.clear()
        return to_return

    def clearFileStreamData(self):
        self.file_stream_data.clear()

    def __update_file_streaming(self):
        """
        Should be called after the packet is validated
        """
        header = ThreespaceHeader.from_bytes(self.com.read(self.header_info.size), self.header_info)
        data = self.com.read(header.length)
        self.file_stream_data += data
        self.file_stream_length -= header.length
        if header.length < 512 or self.file_stream_length == 0: #File streaming sends in chunks of 512. If not 512, it must be the last packet
            self.is_file_streaming = False
            if self.file_stream_length != 0:
                print(f"File streaming stopped due to last packet. However still expected {self.file_stream_length} more bytes.")

    def startDataLogging(self) -> ThreespaceCmdResult[None]:
        self.check_dirty()

        self.header_enabled = True
        self.cache_streaming_settings()

        #Must check whether streaming is being done alongside logging or not. Also configure required settings if it is
        streaming = bool(int(self.get_settings("log_immediate_output")))
        if streaming:
            self.set_settings(log_immediate_output_header_enabled=1,
                                log_immediate_output_header_mode=THREESPACE_OUTPUT_MODE_BINARY) #Must have header enabled in the log messages for this to work and must use binary for the header
        cmd = self.__get_command("__startDataLogging")
        cmd.send_command(self.com, header_enabled=self.header_enabled)
        if self.header_enabled:
            self.__await_command(cmd)
            header = ThreespaceHeader.from_bytes(self.com.read(self.header_info.size), self.header_info)
        else:
            header = ThreespaceHeader()

        self.is_log_streaming = streaming 
        return ThreespaceCmdResult(None, header)

    def stopDataLogging(self) -> ThreespaceCmdResult[None]:
        self.check_dirty()

        cmd = self.__get_command("__stopDataLogging")
        cmd.send_command(self.com, header_enabled=self.header_enabled)

        if self.header_enabled: #Header will be enabled while streaming, but this is useful for startup
            self.__await_command(cmd)
            header = ThreespaceHeader.from_bytes(self.com.read(self.header_info.size), self.header_info)
        else:
            header = ThreespaceHeader()
        #TODO: Remove me now that realignment exists and multiple things can be streaming at once
        if self.is_log_streaming:
            time.sleep(0.05)
            while self.com.length:
                self.com.read_all()

        self.is_log_streaming = False
        return ThreespaceCmdResult(None, header)

    def __update_log_streaming(self):
        """
        Should be called after the packet is validated
        Log streaming is essentially file streaming done as the file is recorded. So uses file
        streaming logistics. Will update this later to also parse the response maybe.
        """
        header = ThreespaceHeader.from_bytes(self.com.read(self.header_info.size), self.header_info)
        data = self.com.read(header.length)
        self.file_stream_data += data

    def softwareReset(self):
        self.check_dirty()
        cmd = self.commands[226]
        cmd.send_command(self.com)
        self.com.close()
        time.sleep(0.5) #Give it time to restart
        self.com.open()
        if self.immediate_debug:
            time.sleep(2) #An additional 2 seconds to ensure can clear all debug messages
            self.com.read_all()
        self.__firmware_init()

    def enterBootloader(self):
        if self.in_bootloader: return

        cmd = self.commands[229]
        cmd.send_command(self.com)
        time.sleep(0.5) #Give it time to boot into bootloader
        if self.com.reenumerates:
            self.com.close()
            success = self.__attempt_rediscover_self()
            if not success:
                raise RuntimeError("Failed to reconnect to sensor in bootloader")
        in_bootloader = self.__check_bootloader_status()
        if not in_bootloader:
            raise RuntimeError("Failed to enter bootloader")
        self.__cached_in_bootloader = True
        self.com.read_all() #Just in case any garbage floating around


    @property
    def in_bootloader(self):
        #This function should not be used internally when solving dirty checks
        self.check_dirty() #If dirty, this we reobtain the value of __cached_in_bootloader.
        return self.__cached_in_bootloader

    def __check_bootloader_status(self):
        """
        Checks if in the bootloader via command. If wanting via cache, just check .in_bootloader
        This function both updates .in_bootloader and returns the value
        
        Must not call this function while streaming. It is only used internally and should be able to meet these conditions
        A user of this class should use .in_bootloader instead of this function        .

        To check, ? is sent, the bootloader will respond with OK. However, to avoid needing to wait
        for the timeout, we send a setting query at the same time. If the response is to the setting, in firmware,
        else if ok, in bootloader. If times out, something funky is happening.
        All bootloader commands are CAPITAL letters. Firmware commands are case insensitive. So as long as send no capitals, its fine.
        """
        #If sending commands over BT to the bootloader, it does an Auto Baudrate Detection
        #for the BT module that requires sending 3 U's. This will respond with 1-2 OK responses if in bootloader.
        #By then adding a ?UUU, that will trigger a <KEY_ERROR> if in firmware. So, can tell if in bootloader or firmware by checking for OK or <KEY_ERROR>
        bootloader = False
        self.com.write("UUU?UUU\n".encode())
        response = self.com.read(2)
        if len(response) == 0: 
            raise RuntimeError("Failed to discover bootloader or firmware. Is the sensor a 3.0?")
        if response == b'OK':
            bootloader = True
        self.com.read_all() #Remove the rest of the OK responses or the rest of the <KEY_ERROR> response
        return bootloader
    
    def bootloader_get_sn(self):
        self.com.write("Q".encode())
        result = self.com.read(9) #9 Because it includes a line feed for reasons
        if len(result) != 9:
            raise Exception()
        #Note bootloader uses big endian instead of little for reasons
        return struct.unpack(f">{_3space_format_to_external('U')}", result[:8])[0]

    def bootloader_boot_firmware(self):
        if not self.in_bootloader: return
        self.com.write("B".encode())
        time.sleep(0.5) #Give time to boot into firmware
        if self.com.reenumerates:
            self.com.close()
            success = self.__attempt_rediscover_self()
            if not success:
                raise RuntimeError("Failed to reconnect to sensor in firmware")
        self.com.read_all() #If debug_mode=1, might be debug messages waiting
        if self.immediate_debug:
            print("Waiting longer before booting into firmware because immediate debug was enabled.")
            time.sleep(2)
            self.com.read_all()
        in_bootloader = self.__check_bootloader_status()
        if in_bootloader:
            raise RuntimeError("Failed to exit bootloader")
        self.__cached_in_bootloader = False
        self.__firmware_init() 
    
    def bootloader_erase_firmware(self, timeout=20):
        """
        This may take a long time
        """
        self.com.write('S'.encode())
        if timeout is not None:
            cached_timeout = self.com.timeout
            self.com.timeout = timeout
        response = self.com.read(1)[0]
        if timeout is not None:
            self.com.timeout = cached_timeout
        return response
    
    def bootloader_get_info(self):
        self.com.write('I'.encode())
        memstart = struct.unpack(f">{_3space_format_to_external('l')}", self.com.read(4))[0]
        memend = struct.unpack(f">{_3space_format_to_external('l')}", self.com.read(4))[0]
        pagesize = struct.unpack(f">{_3space_format_to_external('I')}", self.com.read(2))[0]
        bootversion = struct.unpack(f">{_3space_format_to_external('I')}", self.com.read(2))[0]
        return ThreespaceBootloaderInfo(memstart, memend, pagesize, bootversion)

    def bootloader_prog_mem(self, bytes: bytearray):
        memsize = len(bytes)
        checksum = sum(bytes)
        self.com.write('C'.encode())
        self.com.write(struct.pack(f">{_3space_format_to_external('I')}", memsize))
        self.com.write(bytes)
        self.com.write(struct.pack(f">{_3space_format_to_external('B')}", checksum & 0xFFFF))
        return self.com.read(1)[0]

    def bootloader_get_state(self):
        self.com.write('OO'.encode()) #O is sent twice to compensate for a bug in some versions of the bootloader where the next character is ignored (except for R, do NOT send R after O, it will erase all settings)
        state = struct.unpack(f">{_3space_format_to_external('u')}", self.com.read(4))[0]
        self.com.read_all() #Once the bootloader is fixed, it will respond twice instead of once. So consume any remainder
        return state

    def bootloader_restore_factory_settings(self):
        self.com.write("RR".encode())

    def cleanup(self):
        if not self.in_bootloader:
            if self.is_data_streaming:
                self.stopStreaming()
            if self.is_file_streaming:
                self.fileStopStream()
            if self.is_log_streaming:
                self.stopDataLogging()
            #self.closeFile() #May not be opened, but also not cacheing that so just attempt to close. Currently commented out because breaks embedded
        self.com.close()

#-------------------------START ALL PROTOTYPES------------------------------------

    def eeptsStart(self) -> ThreespaceCmdResult[None]:
        raise NotImplementedError("This method is not available.")

    def eeptsStop(self) -> ThreespaceCmdResult[None]:
        raise NotImplementedError("This method is not available.")    
    
    def eeptsGetOldestStep(self) -> ThreespaceCmdResult[list]:
        raise NotImplementedError("This method is not available.")  

    def eeptsGetNewestStep(self) -> ThreespaceCmdResult[list]:
        raise NotImplementedError("This method is not available.")      

    def eeptsGetNumStepsAvailable(self) -> ThreespaceCmdResult[int]:
        raise NotImplementedError("This method is not available.")       

    def eeptsInsertGPS(self, latitude: float, longitude: float) -> ThreespaceCmdResult[None]:
        raise NotImplementedError("This method is not available.")      
       
    def eeptsAutoOffset(self) -> ThreespaceCmdResult[None]:
        raise NotImplementedError("This method is not available.")    
    
    def getRawGyroRate(self, id: int) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")

    def getRawAccelVec(self, id: int) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")

    def getRawMagVec(self, id: int) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")   

    def getTaredOrientation(self) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.") 

    def getTaredOrientationAsEulerAngles(self) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")  
                                    
    def getTaredOrientationAsRotationMatrix(self) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")

    def getTaredOrientationAsAxisAngles(self) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")

    def getTaredOrientationAsTwoVector(self) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")  
    
    def getDifferenceQuaternion(self) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")   

    def getUntaredOrientation(self) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")  

    def getUntaredOrientationAsEulerAngles(self) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")  

    def getUntaredOrientationAsRotationMatrix(self) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")  

    def getUntaredOrientationAsAxisAngles(self) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")  

    def getUntaredOrientationAsTwoVector(self) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")  
    
    def commitSettings(self) -> ThreespaceCmdResult[None]:
        raise NotImplementedError("This method is not available.")    

    def getMotionlessConfidenceFactor(self) -> ThreespaceCmdResult[float]:
        raise NotImplementedError("This method is not available.")
    
    def enableMSC(self) -> ThreespaceCmdResult[None]:
        raise NotImplementedError("This method is not available.")

    def disableMSC(self) -> ThreespaceCmdResult[None]:
        raise NotImplementedError("This method is not available.")    
    
    def getNextDirectoryItem(self) -> ThreespaceCmdResult[list[int,str,int]]:    
        raise NotImplementedError("This method is not available.")
    
    def changeDirectory(self, path: str) -> ThreespaceCmdResult[None]:    
        raise NotImplementedError("This method is not available.")    

    def openFile(self, path: str) -> ThreespaceCmdResult[None]:    
        raise NotImplementedError("This method is not available.") 
    
    def closeFile(self) -> ThreespaceCmdResult[None]:    
        raise NotImplementedError("This method is not available.")    

    def fileGetRemainingSize(self) -> ThreespaceCmdResult[int]:    
        raise NotImplementedError("This method is not available.")    

    def fileReadLine(self) -> ThreespaceCmdResult[str]:    
        raise NotImplementedError("This method is not available.")     

    def fileReadBytes(self, num_bytes: int) -> ThreespaceCmdResult[bytes]:    
        self.check_dirty()
        cmd = self.commands[THREESPACE_FILE_READ_BYTES_COMMAND_NUM]
        cmd.send_command(self.com, num_bytes, header_enabled=self.header_enabled)
        self.__await_command(cmd)
        if self.header_enabled:
            header = ThreespaceHeader.from_bytes(self.com.read(self.header_info.size), self.header_info)
        else:
            header = ThreespaceHeader()

        response = self.com.read(num_bytes)
        return ThreespaceCmdResult(response, header, data_raw_binary=response)

    def deleteFile(self, path: str) -> ThreespaceCmdResult[None]:    
        raise NotImplementedError("This method is not available.") 

    def getStreamingBatch(self):
        raise NotImplementedError("This method is not available.")

    def setOffsetWithCurrentOrientation(self) -> ThreespaceCmdResult[None]:
        raise NotImplementedError("This method is not available.")
    
    def resetBaseOffset(self) -> ThreespaceCmdResult[None]:
        raise NotImplementedError("This method is not available.")

    def setBaseOffsetWithCurrentOrientation(self) -> ThreespaceCmdResult[None]:
        raise NotImplementedError("This method is not available.")    

    def getTaredTwoVectorInSensorFrame(self) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")    

    def getUntaredTwoVectorInSensorFrame(self) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")        

    def getPrimaryBarometerPressure(self) -> ThreespaceCmdResult[float]:
        raise NotImplementedError("This method is not available.")
    
    def getPrimaryBarometerAltitude(self) -> ThreespaceCmdResult[float]:
        raise NotImplementedError("This method is not available.")    

    def getBarometerAltitude(self, id: int) -> ThreespaceCmdResult[float]:
        raise NotImplementedError("This method is not available.")   
    
    def getBarometerPressure(self, id: int) -> ThreespaceCmdResult[float]:
        raise NotImplementedError("This method is not available.")      

    def getAllPrimaryNormalizedData(self) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")
    
    def getPrimaryNormalizedGyroRate(self) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")

    def getPrimaryNormalizedAccelVec(self) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")

    def getPrimaryNormalizedMagVec(self) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")
    
    def getAllPrimaryCorrectedData(self) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")
    
    def getPrimaryCorrectedGyroRate(self) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")

    def getPrimaryCorrectedAccelVec(self) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")

    def getPrimaryCorrectedMagVec(self) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")    
    
    def getPrimaryGlobalLinearAccel(self) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.") 

    def getPrimaryLocalLinearAccel(self) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")         

    def getTemperatureCelsius(self) -> ThreespaceCmdResult[float]:
        raise NotImplementedError("This method is not available.") 
    
    def getTemperatureFahrenheit(self) -> ThreespaceCmdResult[float]:
        raise NotImplementedError("This method is not available.")     

    def getNormalizedGyroRate(self, id: int) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")
    
    def getNormalizedAccelVec(self, id: int) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")

    def getNormalizedMagVec(self, id: int) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")        

    def getCorrectedGyroRate(self, id: int) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")
    
    def getCorrectedAccelVec(self, id: int) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")

    def getCorrectedMagVec(self, id: int) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")        

    def enableMSC(self) -> ThreespaceCmdResult[None]:
        raise NotImplementedError("This method is not available.")  
    
    def disableMSC(self) -> ThreespaceCmdResult[None]:
        raise NotImplementedError("This method is not available.")    

    def getTimestamp(self) -> ThreespaceCmdResult[int]:
        raise NotImplementedError("This method is not available.")  

    def getBatteryVoltage(self) -> ThreespaceCmdResult[float]:
        raise NotImplementedError("This method is not available.") 
    
    def getBatteryPercent(self) -> ThreespaceCmdResult[int]:
        raise NotImplementedError("This method is not available.") 

    def getBatteryStatus(self) -> ThreespaceCmdResult[int]:
        raise NotImplementedError("This method is not available.")         

    def getGpsCoord(self) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")
    
    def getGpsAltitude(self) -> ThreespaceCmdResult[float]:
        raise NotImplementedError("This method is not available.") 

    def getGpsFixState(self) -> ThreespaceCmdResult[int]:
        raise NotImplementedError("This method is not available.") 

    def getGpsHdop(self) -> ThreespaceCmdResult[float]:
        raise NotImplementedError("This method is not available.") 

    def getGpsSatellites(self) -> ThreespaceCmdResult[int]:
        raise NotImplementedError("This method is not available.")                 

    def getButtonState(self) -> ThreespaceCmdResult[int]:
        raise NotImplementedError("This method is not available.")

    def correctRawGyroData(self, x: float, y: float, z: float, id: int) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")
    
    def correctRawAccelData(self, x: float, y: float, z: float, id: int) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")

    def correctRawMagData(self, x: float, y: float, z: float, id: int) -> ThreespaceCmdResult[list[float]]:
        raise NotImplementedError("This method is not available.")

    def formatSd(self) -> ThreespaceCmdResult[None]:
        raise NotImplementedError("This method is not available.")    

    def setDateTime(self, year: int, month: int, day: int, hour: int, minute: int, second: int) -> ThreespaceCmdResult[None]:
        raise NotImplementedError("This method is not available.")

    def getDateTime(self) -> ThreespaceCmdResult[list[int]]:
        raise NotImplementedError("This method is not available.")  

    def tareWithCurrentOrientation(self) -> ThreespaceCmdResult[None]:
        raise NotImplementedError("This method is not available.") 
    
    def setBaseTareWithCurrentOrientation(self) -> ThreespaceCmdResult[None]:
        raise NotImplementedError("This method is not available.")     

    def resetFilter(self) -> ThreespaceCmdResult[None]:
        raise NotImplementedError("This method is not available.")
    
    def getNumDebugMessages(self) -> ThreespaceCmdResult[int]:
        raise NotImplementedError("This method is not available.")
    
    def getOldestDebugMessage(self) -> ThreespaceCmdResult[str]:
        raise NotImplementedError("This method is not available.")
    
    def beginPassiveAutoCalibration(self, enabled_bitfield: int) -> ThreespaceCmdResult[None]:
        raise NotImplementedError("This method is not available.")

    def getActivePassiveAutoCalibration(self) -> ThreespaceCmdResult[int]:
        raise NotImplementedError("This method is not available.")

    def beginActiveAutoCalibration(self) -> ThreespaceCmdResult[None]:
        raise NotImplementedError("This method is not available.")

    def isActiveAutoCalibrationActive(self) -> ThreespaceCmdResult[int]:
        raise NotImplementedError("This method is not available.")                
    
    def getStreamingLabel(self, cmd_num: int) -> ThreespaceCmdResult[str]:
        raise NotImplementedError("This method is not available.")   

    def setCursor(self, cursor_index: int) -> ThreespaceCmdResult[None]:
        raise NotImplementedError("This method is not available.") 

    def getLastLogCursorInfo(self) -> ThreespaceCmdResult[tuple[int,str]]:
        raise NotImplementedError("This method is not available.")   

    def pauseLogStreaming(self, pause: bool) -> ThreespaceCmdResult[None]:
        raise NotImplementedError("This method is not available.")               

THREESPACE_GET_STREAMING_BATCH_COMMAND_NUM = 84
THREESPACE_FILE_READ_BYTES_COMMAND_NUM = 177

#Acutal command definitions
_threespace_commands: list[ThreespaceCommand] = [
    #Tared Orientation
    ThreespaceCommand("getTaredOrientation", 0, "", "ffff"),
    ThreespaceCommand("getTaredOrientationAsEulerAngles", 1, "", "fff"),
    ThreespaceCommand("getTaredOrientationAsRotationMatrix", 2, "", "fffffffff"),
    ThreespaceCommand("getTaredOrientationAsAxisAngles", 3, "", "ffff"),
    ThreespaceCommand("getTaredOrientationAsTwoVector", 4, "", "ffffff"),

    #Weird
    ThreespaceCommand("getDifferenceQuaternion", 5, "", "ffff"),

    #Untared Orientation
    ThreespaceCommand("getUntaredOrientation", 6, "", "ffff"),
    ThreespaceCommand("getUntaredOrientationAsEulerAngles", 7, "", "fff"),
    ThreespaceCommand("getUntaredOrientationAsRotationMatrix", 8, "", "fffffffff"),
    ThreespaceCommand("getUntaredOrientationAsAxisAngles", 9, "", "ffff"),
    ThreespaceCommand("getUntaredOrientationAsTwoVector", 10, "", "ffffff"),
    
    #Late orientation additions
    ThreespaceCommand("getTaredTwoVectorInSensorFrame", 11, "", "ffffff"),
    ThreespaceCommand("getUntaredTwoVectorInSensorFrame", 12, "", "ffffff"),

    ThreespaceCommand("getPrimaryBarometerPressure", 13, "", "f"),
    ThreespaceCommand("getPrimaryBarometerAltitude", 14, "", "f"),
    ThreespaceCommand("getBarometerAltitude", 15, "b", "f"),
    ThreespaceCommand("getBarometerPressure", 16, "b", "f"),

    ThreespaceCommand("setOffsetWithCurrentOrientation", 19, "", ""),
    ThreespaceCommand("resetBaseOffset", 20, "", ""),
    ThreespaceCommand("setBaseOffsetWithCurrentOrientation", 22, "", ""),

    ThreespaceCommand("getAllPrimaryNormalizedData", 32, "", "fffffffff"),
    ThreespaceCommand("getPrimaryNormalizedGyroRate", 33, "", "fff"),
    ThreespaceCommand("getPrimaryNormalizedAccelVec", 34, "", "fff"),
    ThreespaceCommand("getPrimaryNormalizedMagVec", 35, "", "fff"),

    ThreespaceCommand("getAllPrimaryCorrectedData", 37, "", "fffffffff"),
    ThreespaceCommand("getPrimaryCorrectedGyroRate", 38, "", "fff"),
    ThreespaceCommand("getPrimaryCorrectedAccelVec", 39, "", "fff"),
    ThreespaceCommand("getPrimaryCorrectedMagVec", 40, "", "fff"),

    ThreespaceCommand("getPrimaryGlobalLinearAccel", 41, "", "fff"),
    ThreespaceCommand("getPrimaryLocalLinearAccel", 42, "", "fff"),

    ThreespaceCommand("getTemperatureCelsius", 43, "", "f"),
    ThreespaceCommand("getTemperatureFahrenheit", 44, "", "f"),

    ThreespaceCommand("getMotionlessConfidenceFactor", 45, "", "f"),

    ThreespaceCommand("correctRawGyroData", 48, "fffb", "fff"),
    ThreespaceCommand("correctRawAccelData", 49, "fffb", "fff"),
    ThreespaceCommand("correctRawMagData", 50, "fffb", "fff"),

    ThreespaceCommand("getNormalizedGyroRate", 51, "b", "fff"),
    ThreespaceCommand("getNormalizedAccelVec", 52, "b", "fff"),
    ThreespaceCommand("getNormalizedMagVec", 53, "b", "fff"),

    ThreespaceCommand("getCorrectedGyroRate", 54, "b", "fff"),
    ThreespaceCommand("getCorrectedAccelVec", 55, "b", "fff"),
    ThreespaceCommand("getCorrectedMagVec", 56, "b", "fff"),

    ThreespaceCommand("enableMSC", 57, "", ""),
    ThreespaceCommand("disableMSC", 58, "", ""),

    ThreespaceCommand("formatSd", 59, "", ""),
    ThreespaceCommand("__startDataLogging", 60, "", ""),
    ThreespaceCommand("__stopDataLogging", 61, "", ""),

    ThreespaceCommand("setDateTime", 62, "Bbbbbb", ""),
    ThreespaceCommand("getDateTime", 63, "", "Bbbbbb"),

    ThreespaceCommand("getRawGyroRate", 65, "b", "fff"),
    ThreespaceCommand("getRawAccelVec", 66, "b", "fff"),
    ThreespaceCommand("getRawMagVec", 67, "b", "fff"),

    ThreespaceCommand("eeptsStart", 68, "", ""),
    ThreespaceCommand("eeptsStop", 69, "", ""),
    ThreespaceCommand("eeptsGetOldestStep", 70, "", "uuddffffffbbff"),
    ThreespaceCommand("eeptsGetNewestStep", 71, "", "uuddffffffbbff"),
    ThreespaceCommand("eeptsGetNumStepsAvailable", 72, "", "b"),
    ThreespaceCommand("eeptsInsertGPS", 73, "dd", ""),
    ThreespaceCommand("eeptsAutoOffset", 74, "", ""),

    ThreespaceCommand("getStreamingLabel", 83, "b", "S"),
    ThreespaceCommand("__getStreamingBatch", THREESPACE_GET_STREAMING_BATCH_COMMAND_NUM, "", "S"),
    ThreespaceCommand("__startStreaming", 85, "", ""),
    ThreespaceCommand("__stopStreaming", 86, "", ""),
    ThreespaceCommand("pauseLogStreaming", 87, "b", ""),
    
    ThreespaceCommand("getTimestamp", 94, "", "U"),

    ThreespaceCommand("tareWithCurrentOrientation", 96, "", ""),
    ThreespaceCommand("setBaseTareWithCurrentOrientation", 97, "", ""),

    ThreespaceCommand("resetFilter", 120, "", ""),
    ThreespaceCommand("getNumDebugMessages", 126, "", "B"),
    ThreespaceCommand("getOldestDebugMessage", 127, "", "S"),

    ThreespaceCommand("beginPassiveAutoCalibration", 165, "b", ""),
    ThreespaceCommand("getActivePassiveAutoCalibration", 166, "", "b"),
    ThreespaceCommand("beginActiveAutoCalibration", 167, "", ""),
    ThreespaceCommand("isActiveAutoCalibrationActive", 168, "", "b"),

    ThreespaceCommand("getLastLogCursorInfo", 170, "", "US"),
    ThreespaceCommand("getNextDirectoryItem", 171, "", "bsU"),
    ThreespaceCommand("changeDirectory", 172, "S", ""),
    ThreespaceCommand("openFile", 173, "S", ""),
    ThreespaceCommand("closeFile", 174, "", ""),
    ThreespaceCommand("fileGetRemainingSize", 175, "", "U"),
    ThreespaceCommand("fileReadLine", 176, "", "S"),
    ThreespaceCommand("__fileReadBytes", THREESPACE_FILE_READ_BYTES_COMMAND_NUM, "B", "S"), #This has to be handled specially as the output is variable length BYTES not STRING
    ThreespaceCommand("deleteFile", 178, "S", ""),
    ThreespaceCommand("setCursor", 179, "U", ""),
    ThreespaceCommand("__fileStartStream", 180, "", "U"),
    ThreespaceCommand("__fileStopStream", 181, "", ""),

    ThreespaceCommand("getBatteryVoltage", 201, "", "f"),
    ThreespaceCommand("getBatteryPercent", 202, "", "b"),
    ThreespaceCommand("getBatteryStatus", 203, "", "b"),

    ThreespaceCommand("getGpsCoord", 215, "", "dd"),
    ThreespaceCommand("getGpsAltitude", 216, "", "f"),
    ThreespaceCommand("getGpsFixState", 217, "", "b"),
    ThreespaceCommand("getGpsHdop", 218, "", "f"),
    ThreespaceCommand("getGpsSatellites", 219, "", "b"),

    ThreespaceCommand("commitSettings", 225, "", ""),
    ThreespaceCommand("__softwareReset", 226, "", ""),
    ThreespaceCommand("__enterBootloader", 229, "", ""),

    ThreespaceCommand("getButtonState", 250, "", "b"),
]

def threespaceCommandGet(cmd_num: int):
    for command in _threespace_commands:
        if command.info.num == cmd_num:
            return command
    return None

def threespaceCommandGetInfo(cmd_num: int):
    command = threespaceCommandGet(cmd_num)
    if command is None: return None
    return command.info

def threespaceGetHeaderLabels(header_info: ThreespaceHeaderInfo):
    order = []
    if header_info.status_enabled:
        order.append("status")
    if header_info.timestamp_enabled:
        order.append("timestamp")
    if header_info.echo_enabled:
        order.append("echo")
    if header_info.checksum_enabled:
        order.append("checksum")
    if header_info.serial_enabled:
        order.append("serial#")
    if header_info.length_enabled:
        order.append("len")
    return order


def vec_len(vector: list[float|int]):
    return sum(v ** 2 for v in vector) ** 0.5

def vec_dot(a: list[float], b: list[float]):
    return sum(a[i] * b[i] for i in range(len(a)))

def vec_cross(a: list[float], b: list[float]):
    cross = [0, 0, 0]
    cross[0] = a[1] * b[2] - a[2] * b[1]
    cross[1] = a[2] * b[0] - a[0] * b[2]
    cross[2] = a[0] * b[1] - a[1] * b[0]
    return cross

def vec_normalize(vec: list[float]):
    l = vec_len(vec)
    if l == 0:
        return vec
    return [v / l for v in vec]

def vec_is_right_handed(order: str, negations: list[bool]):
    right_handed = order.lower() in ("xzy", "yxz", "zyx")
    for i in range(3):
        if negations[i]:
            right_handed = not right_handed
    return right_handed

def axis_to_unit_vector(axis: str):
    axis = axis.lower()
    if axis == 'x' or axis == 0: return [1, 0, 0]
    if axis == 'y' or axis == 1: return [0, 1, 0]
    if axis == 'z' or axis == 2: return [0, 0, 1]


from yostlabs.communication.base import *
import serial
import serial.tools.list_ports


class ThreespaceSerialComClass(ThreespaceComClass):

    PID_V3_MASK = 0x3000
    PID_BOOTLOADER = 0x1000

    VID = 0x2476

    DEFAULT_BAUDRATE = 115200
    DEFAULT_TIMEOUT = 2

    def __init__(self, ser: serial.Serial | str):
        if isinstance(ser, serial.Serial):
            self.ser = ser
        elif isinstance(ser, str):
            self.ser = serial.Serial(ser, baudrate=ThreespaceSerialComClass.DEFAULT_BAUDRATE, timeout=ThreespaceSerialComClass.DEFAULT_TIMEOUT)
        else:
            raise TypeError("Invalid type for creating a ThreespaceSerialComClass:", type(ser), ser)

        self.ser = ser

        self.peek_buffer = bytearray()
        self.peek_length = 0

    def write(self, bytes: bytes):
        self.ser.write(bytes)
    
    def read(self, num_bytes: int):
        if self.peek_length >= num_bytes:
            result = self.peek_buffer[:num_bytes]
            self.peek_length -= num_bytes
            del self.peek_buffer[:num_bytes]
        else:
            result = self.peek_buffer + self.ser.read(num_bytes - self.peek_length) #Must supply the amount desired to read instead of just buffering so the timeout works
            self.peek_buffer.clear()
            self.peek_length = 0
        return result
    
    def peek(self, num_bytes: int):
        if self.peek_length >= num_bytes:
            return self.peek_buffer[:num_bytes]
        else:
            self.peek_buffer += self.ser.read(num_bytes - self.peek_length) #Must supply the amount desired to read instead of just buffering so the timeout works
            self.peek_length = len(self.peek_buffer) #The read may have timed out, so calculate new size
            return self.peek_buffer.copy()
        
    def read_until(self, expected: bytes) -> bytes:
        if expected in self.peek_buffer:
            length = self.peek_buffer.index(expected) + len(expected)
            result = self.peek_buffer[:length]
            self.peek_length -= length
            del self.peek_buffer[:length]
            return result
        #Have to actually read from serial port until the data is available
        result = self.peek_buffer + self.ser.read_until(expected)
        self.peek_buffer.clear()
        self.peek_length = 0
        return result

    def peek_until(self, expected: bytes, max_length: int = None) -> bytes:
        if expected in self.peek_buffer:
            length = self.peek_buffer.index(expected) + len(expected)
            if max_length is not None and length > max_length: 
                length = max_length
            result = self.peek_buffer[:length]
            return result
        if max_length is not None and self.peek_length >= max_length:
            return self.peek_buffer[:max_length]
        
        #Have to actually read from serial port until the data is available
        if max_length is not None:
            max_length = max(0, max_length - self.peek_length)
        self.peek_buffer += self.ser.read_until(expected, size=max_length)
        self.peek_length = len(self.peek_buffer)
        return self.peek_buffer.copy()

    def close(self):
        self.ser.close()
        self.peek_buffer.clear()
        self.peek_length = 0
    
    def open(self):
        try:
            self.ser.open()
        except:
            return False
        return True

    def check_open(self):
        try:
            self.ser.in_waiting
        except:
            return False
        return True

    @property
    def length(self):
        return self.peek_length + self.ser.in_waiting    

    @property
    def timeout(self) -> float:
        return self.ser.timeout
    
    @timeout.setter
    def timeout(self, timeout: float):
        self.ser.timeout = timeout    

    @property
    def reenumerates(self) -> bool:
        return True
    
    @property
    def name(self) -> str:
        return self.ser.port

    #This is not part of the ThreespaceComClass interface, but is useful as a utility for those directly using the ThreespaceSerialComClass
    @staticmethod 
    def enumerate_ports():
        cls = ThreespaceSerialComClass
        ports = serial.tools.list_ports.comports()
        for port in ports:
            if port.vid == cls.VID and (port.pid & cls.PID_V3_MASK == cls.PID_V3_MASK or port.pid == cls.PID_BOOTLOADER): 
                yield port

    @staticmethod
    def auto_detect(default_timeout: float = 2, default_baudrate: int = 115200) -> Generator["ThreespaceSerialComClass", None, None]:
        """
        Returns a list of com classes of the same type called on nearby.
        These ports will start unopened. This allows the caller to get a list of ports without having to connect.
        """
        cls = ThreespaceSerialComClass
        ports = serial.tools.list_ports.comports()
        for port in ports:
            if port.vid == cls.VID and (port.pid & cls.PID_V3_MASK == cls.PID_V3_MASK or port.pid == cls.PID_BOOTLOADER): 
                ser = serial.Serial(None, baudrate=default_baudrate, timeout=default_timeout) #By setting port as None, can create an object without immediately opening the port
                ser.port = port.device #Now assign the port, allowing the serial object to exist without being opened yet
                yield ThreespaceSerialComClass(ser)


THREESPACE_HEADER_STATUS_BIT_POS = 0
THREESPACE_HEADER_TIMESTAMP_BIT_POS = 1
THREESPACE_HEADER_ECHO_BIT_POS = 2
THREESPACE_HEADER_CHECKSUM_BIT_POS = 3
THREESPACE_HEADER_SERIAL_BIT_POS = 4
THREESPACE_HEADER_LENGTH_BIT_POS = 5
THREESPACE_HEADER_NUM_BITS = 6

THREESPACE_HEADER_STATUS_BIT = (1 << THREESPACE_HEADER_STATUS_BIT_POS)
THREESPACE_HEADER_TIMESTAMP_BIT = (1 << THREESPACE_HEADER_TIMESTAMP_BIT_POS)
THREESPACE_HEADER_ECHO_BIT = (1 << THREESPACE_HEADER_ECHO_BIT_POS)
THREESPACE_HEADER_CHECKSUM_BIT = (1 << THREESPACE_HEADER_CHECKSUM_BIT_POS)
THREESPACE_HEADER_SERIAL_BIT = (1 << THREESPACE_HEADER_SERIAL_BIT_POS)
THREESPACE_HEADER_LENGTH_BIT = (1 << THREESPACE_HEADER_LENGTH_BIT_POS)

FIRMWARE_VALID_BIT = 0x01

PASSIVE_CALIBRATE_GYRO = (1 << 0)
PASSIVE_CALIBRATE_MAG_REF = (1 << 1)

STREAMING_MAX_HZ = 2000

THREESPACE_OUTPUT_MODE_ASCII = 1
THREESPACE_OUTPUT_MODE_BINARY = 2


from yostlabs.tss3.api import ThreespaceSensor, StreamableCommands
from yostlabs.communication.serial import ThreespaceSerialComClass
from yostlabs.tss3.utils.parser import ThreespaceBinaryParser
import time

#Create a sensor by auto detecting a ThreespaceSerialComClass
sensor = ThreespaceSensor(ThreespaceSerialComClass)

#First, gather some data and store it in a file for this example
sensor.set_settings(stream_slots=f"{StreamableCommands.GetTimestamp.value},{StreamableCommands.GetAllPrimaryCorrectedData.value}")
sensor.set_settings(stream_hz=500)

print("Gathering data for 5 seconds...")
sensor.startStreaming()
start_time = time.time()
FNAME = "logged_data.bin"
file = open(FNAME, 'wb')
while time.time() - start_time < 5:
    sensor.updateStreaming()
    packet = sensor.getOldestStreamingPacket()
    while packet is not None:
        file.write(packet.raw_binary) #Store the raw binary for parsing later
        packet = sensor.getOldestStreamingPacket()
sensor.stopStreaming()

file.close()

#Create the parser
parser = ThreespaceBinaryParser(verbose=True)

#Configure it to have the same settings used so it knows how to read the data
parser.set_header(sensor.header_info)

#Register all command responses that could have been included. This includes more then streaming data,
#however because this example did streaming, it is going to register the getStreamingBatch command to parse.
#Some commands take additional KWARGS to know how to parse. Ex: getStreamingBatch requires knowing what the streaming slots were
parser.register_command(84, stream_slots=sensor.streaming_slots)

#Once all commands are registered, simply feed the parser data and attempt to retrieve messages
file = open(FNAME, "rb")

print("Parsing the gathered data from the binary file")
parser.insert_data(file.read())
msg = parser.parse_message()
while msg is not None:
    print(msg)
    msg = parser.parse_message()

file.close()